@@ -1,10 +1,15 @@
 from hashlib import md5
 import io
+import os
+import logging
 from .utils import read_block, tokenize
+logger = logging.getLogger('fsspec')
 
 # alternative names for some methods, which get patched to new instances
+# (alias, original)
 aliases = [
     ('makedir', 'mkdir'),
+    ('mkdirs', 'makedirs'),
     ('listdir', 'ls'),
     ('cp', 'copy'),
     ('move', 'mv'),
@@ -12,17 +17,25 @@ aliases = [
     ('disk_usage', 'du'),
     ('rename', 'mv'),
     ('delete', 'rm'),
+    ('upload', 'put'),
+    ('download', 'get')
 ]
 
+try:   # optionally derive from pyarrow's FileSystem, if available
+    import pyarrow as pa
+    up = pa.filesystem.DaskFileSystem
+except ImportError:
+    up = object
 
-class AbstractFileSystem(object):
+
+class AbstractFileSystem(up):
     """
     An abstract super-class for pythonic file-systems
 
     Implementations are expected to be compatible with or, better, subclass
     from here.
     """
-    _singleton = None
+    _singleton = [None]  # will contain the newest instance
     _cache = None
     cachable = True  # this class can be cached, instances reused
     _cached = False
@@ -38,9 +51,6 @@ class AbstractFileSystem(object):
 
         The instance will skip init if instance.cached = True.
         """
-        if cls._singleton is None:
-            # set up space for singleton
-            cls._singleton = [None]
         if cls._cache is None and cls.cachable:
             # set up instance cache, if using
             cls._cache = {}
@@ -81,11 +91,29 @@ class AbstractFileSystem(object):
         self._transaction = Transaction(self)
         self._singleton[0] = self
         self.dircache = {}
-        if storage_options.get('add_docs', True):
+        if storage_options.pop('add_docs', True):
             self._mangle_docstrings()
-        if storage_options.get('add_aliases', True):
+        if storage_options.pop('add_aliases', True):
             for new, old in aliases:
-                setattr(self, new, getattr(self, old))
+                if not hasattr(self, new):
+                    # don't apply alias if attribute exists already
+                    setattr(self, new, getattr(self, old))
+
+    def __dask_tokenize__(self):
+        return self.token
+
+    def __hash__(self):
+        return int(self._tok, 16)
+
+    def __eq__(self, other):
+        return self.token == other.token
+
+    @classmethod
+    def clear_instance_cache(cls, remove_singleton=True):
+        """Remove any instances stored in class attributes"""
+        cls._cache.clear()
+        if remove_singleton:
+            cls._singleton = [None]
 
     def _mangle_docstrings(self):
         """Add AbstractFileSystem docstrings to subclass methods
@@ -119,14 +147,16 @@ class AbstractFileSystem(object):
 
         May require FS-specific handling, e.g., for relative paths or links.
         """
-        if path.startswith(cls.protocol + '://'):
-            return path[len(cls.protocol) + 3:]
-        elif path.startswith(cls.protocol + ':'):
-            return path[len(cls.protocol) + 1:]
-        elif path.startswith(cls.protocol):
-            return path[len(cls.protocol):]
-        else:
-            return path
+        protos = (cls.protocol, ) if isinstance(
+            cls.protocol, str) else cls.protocol
+        for protocol in protos:
+            path = path.rstrip('/')
+            if path.startswith(protocol + '://'):
+                path = path[len(protocol) + 3:]
+            elif path.startswith(protocol + ':'):
+                path = path[len(protocol) + 1:]
+        # use of root_marker to make minimum required path, e.g., "/"
+        return path or cls.root_marker
 
     @staticmethod
     def _get_kwargs_from_urls(paths):
@@ -218,7 +248,7 @@ class AbstractFileSystem(object):
         """Remove a directory, if empty"""
         pass  # not necessary to implement, may not have directories
 
-    def ls(self, path, detail=False):
+    def ls(self, path, **kwargs):
         """List objects at path.
 
         This should include subdirectories and files at that location. The
@@ -228,13 +258,18 @@ class AbstractFileSystem(object):
         The specific keys, or perhaps a FileInfo class, or similar, is TBD,
         but must be consistent across implementations.
         Must include:
-        - full path to the entry
+        - full path to the entry (without protocol)
         - size of the entry, in bytes
+        - type of entry, "file", "directory" or other
 
         Additional information
-        may be present, approriate to the file-system, e.g., generation,
+        may be present, aproriate to the file-system, e.g., generation,
         checksum, etc.
 
+        May use refresh=True|False to allow use of self._ls_from_cache to
+        check for a saved listing and avoid calling the backend. This would be
+        common where listing may be expensive.
+
         Parameters
         ----------
         path: str
@@ -242,9 +277,32 @@ class AbstractFileSystem(object):
             if True, gives a list of dictionaries, where each is the same as
             the result of ``info(path)``. If False, gives a list of paths
             (str).
+        kwargs: may have additional backend-specific options, such as version
+            information
+
+        Returns
+        -------
+        List of strings if detail is False, or list of directory information
+        dicts if detail is True.
         """
         raise NotImplementedError
 
+    def _ls_from_cache(self, path):
+        """Check cache for listing
+
+        Returns listing, if found (may me empty list for a directly that exists
+        but contains nothing), None if not in cache.
+        """
+        parent = self._parent(path)
+        if path in self.dircache:
+            return self.dircache[path]
+        elif parent in self.dircache:
+            files = [f for f in self.dircache[parent] if f['name'] == path]
+            if len(files) == 0:
+                # parent dir was listed but did not contain this file
+                raise FileNotFoundError(path)
+            return files
+
     def walk(self, path, maxdepth=3):
         """ Return all files belows path
 
@@ -267,13 +325,22 @@ class AbstractFileSystem(object):
         dirs = []
         files = []
 
-        for info in self.ls(path, True):
+        try:
+            listing = self.ls(path, True)
+        except FileNotFoundError:
+            return [], [], []
+
+        for info in listing:
             # each info name must be at least [path]/part , but here
             # we check also for names like [path]/part/
-            name = info['name']
-            if info['type'] == 'directory':
-                full_dirs.append(name.rstrip('/'))
-                dirs.append(name.rstrip('/').rsplit('/', 1)[-1])
+            name = info['name'].rstrip('/')
+            if info['type'] == 'directory' and name != path:
+                # do not include "self" path
+                full_dirs.append(name)
+                dirs.append(name.rsplit('/', 1)[-1])
+            elif name == path:
+                # file-like with same name as give path
+                files.append('')
             else:
                 files.append(name.rsplit('/', 1)[-1])
         yield path, dirs, files
@@ -289,10 +356,17 @@ class AbstractFileSystem(object):
 
         Like posix ``find`` command without conditions
         """
+        # TODO: allow equivalent of -name parameter
         out = []
         for path, _, files in self.walk(path, maxdepth):
             for name in files:
-                out.append('/'.join([path.rstrip('/'), name]) if path else name)
+                if name:
+                    out.append('/'.join([path.rstrip('/'), name])
+                               if path else name)
+        if self.isfile(path) and path not in out:
+            # walk works on directories, but find should also return [path]
+            # when path happens to be a file
+            out.append(path)
         return out
 
     def du(self, path, total=True, maxdepth=4):
@@ -300,6 +374,7 @@ class AbstractFileSystem(object):
 
         Parameters
         ----------
+        path: str
         total: bool
             whether to sum all the file sizes
         maxdepth: int or None
@@ -332,28 +407,32 @@ class AbstractFileSystem(object):
         Example reimplements code in ``glob.glob()``, taken from hdfs3.
         """
         import re
-        import posixpath
-        if "*" not in path:
+        ends = path.endswith('/')
+        path = self._strip_protocol(path)
+        indstar = path.find("*") if path.find("*") >= 0 else len(path)
+        indques = path.find("?") if path.find("?") >= 0 else len(path)
+        ind = min(indstar, indques)
+        if "*" not in path and "?" not in path:
             root = path
             depth = 1
-            if path.endswith('/'):
-                path += '*'
+            if ends:
+                path += '/*'
             elif self.exists(path):
                 return [path]
             else:
                 raise FileNotFoundError(path)
-        elif '/' in path[:path.index('*')]:
-            ind = path[:path.index('*')].rindex('/')
-            root = path[:ind + 1]
-            depth = path[ind + 1:].count('/') + 1
+        elif '/' in path[:ind]:
+            ind2 = path[:ind].rindex('/')
+            root = path[:ind2 + 1]
+            depth = 20 if "**" in path else path[ind2 + 1:].count('/') + 1
         else:
             root = ''
-            depth = 1
-        allpaths = []
-        for dirname, dirs, fils in self.walk(root, maxdepth=depth):
-            allpaths.extend(posixpath.join(dirname, f) for f in fils)
-        pattern = re.compile("^" + path.replace('//', '/')
+            depth = 20 if "**" in path else 1
+        allpaths = self.find(root, maxdepth=depth)
+        pattern = re.compile("^" + path.replace('.', r'\.')
+                             .replace('//', '/')
                              .rstrip('/')
+                             .replace('**', '.+')
                              .replace('*', '[^/]*')
                              .replace('?', '.') + "$")
         out = {p for p in allpaths
@@ -368,30 +447,47 @@ class AbstractFileSystem(object):
         except:   # any exception allowed bar FileNotFoundError?
             return False
 
-    def info(self, path):
+    def info(self, path, **kwargs):
         """Give details of entry at path
 
         Returns a single dictionary, with exactly the same information as ``ls``
         would with ``detail=True``.
 
-        The default implementation should probably be overridden by a shortcut.
+        The default implementation should calls ls and could be overridden by a
+        shortcut. kwargs are passed on to ```ls()``.
 
         Returns
         -------
         dict with keys: name (full path in the FS), size (in bytes), type (file,
         directory, or something else) and other FS-specific keys.
         """
-        out = self.ls(path, detail=True)
+        path = self._strip_protocol(path)
+        out = self.ls(self._parent(path), detail=True, **kwargs)
+        out = [o for o in out if o['name'].rstrip('/') == path]
+        if out:
+            return out[0]
+        out = self.ls(path, detail=True, **kwargs)
         path = path.rstrip('/')
-        out = [o for o in out if o['name'].rstrip('/') == path]
-        if out:
-            return out[0]
+        out1 = [o for o in out if o['name'].rstrip('/') == path]
+        if len(out1) == 1:
+            return out1[0]
+        elif len(out1) > 1 or out:
+            return {'name': path, 'size': 0, 'type': 'directory'}
+        else:
+            raise FileNotFoundError(path)
 
-        out = self.ls(self._parent(path), detail=True)
-        out = [o for o in out if o['name'].rstrip('/') == path]
-        if out:
-            return out[0]
-        raise FileNotFoundError(path)
+    def checksum(self, path):
+        """Unique value for current version of file
+
+        If the checksum is the same from one moment to another, the contents
+        are guaranteed to be the same. If the checksum changes, the contents
+        *might* have changed.
+
+        This should normally be overridden; default will probably capture
+        creation/modification timestamp (which would be good) or maybe
+        access timestamp (which would be bad)
+        """
+        return int(tokenize(self.info(path)), 16)
 
     def size(self, path):
         """Size in bytes of file"""
@@ -399,37 +495,82 @@ class AbstractFileSystem(object):
 
     def isdir(self, path):
         """Is this entry directory-like?"""
-        return self.info(path)['type'] == 'directory'
+        try:
+            return self.info(path)['type'] == 'directory'
+        except FileNotFoundError:
+            return False
 
     def isfile(self, path):
         """Is this entry file-like?"""
-        return self.info(path)['type'] == 'file'
+        try:
+            return self.info(path)['type'] == 'file'
+        except:
+            return False
 
     def cat(self, path):
         """ Get the content of a file """
         return self.open(path, 'rb').read()
 
-    def get(self, rpath, lpath, **kwargs):
+    def get(self, rpath, lpath, recursive=False, **kwargs):
         """ Copy file to local
 
         Possible extension: maybe should be able to copy to any file-system
         (streaming through local).
         """
-        with self.open(rpath, 'rb') as f1:
-            with open(lpath, 'wb') as f2:
-                data = True
-                while data:
-                    data = f1.read(self.blocksize)
-                    f2.write(data)
+        rpath = self._strip_protocol(rpath)
+        if recursive:
+            rpaths = self.find(rpath)
+            rootdir = os.path.basename(rpath.rstrip('/'))
+            if os.path.isdir(lpath):
+                # copy rpath inside lpath directory
+                lpath2 = os.path.join(lpath, rootdir)
+            else:
+                # copy rpath as lpath directory
+                lpath2 = lpath
+            lpaths = [os.path.join(lpath2, path[len(rpath):].lstrip('/'))
+                      for path in rpaths]
+            for lpath in lpaths:
+                dirname = os.path.dirname(lpath)
+                if not os.path.isdir(dirname):
+                    os.makedirs(dirname)
+        else:
+            rpaths = [rpath]
+            lpaths = [lpath]
+        for lpath, rpath in zip(lpaths, rpaths):
+            print(lpath, rpath)
+            with self.open(rpath, 'rb') as f1:
+                with open(lpath, 'wb') as f2:
+                    data = True
+                    while data:
+                        data = f1.read(self.blocksize)
+                        f2.write(data)
 
-    def put(self, lpath, rpath, **kwargs):
+    def put(self, lpath, rpath, recursive=False, **kwargs):
         """ Upload file from local """
-        with open(lpath, 'rb') as f1:
-            with self.open(rpath, 'wb') as f2:
-                data = True
-                while data:
-                    data = f1.read(self.blocksize)
-                    f2.write(data)
+        if recursive:
+            lpaths = []
+            for dirname, subdirlist, filelist in os.walk(lpath):
+                lpaths += [os.path.join(dirname, filename)
+                           for filename in filelist]
+            rootdir = os.path.basename(lpath.rstrip('/'))
+            if self.exists(rpath):
+                # copy lpath inside rpath directory
+                rpath2 = os.path.join(rpath, rootdir)
+            else:
+                # copy lpath as rpath directory
+                rpath2 = rpath
+            rpaths = [os.path.join(rpath2, path[len(lpath):].lstrip('/'))
+                      for path in lpaths]
+        else:
+            lpaths = [lpath]
+            rpaths = [rpath]
+        for lpath, rpath in zip(lpaths, rpaths):
+            with open(lpath, 'rb') as f1:
+                with self.open(rpath, 'wb') as f2:
+                    data = True
+                    while data:
+                        data = f1.read(self.blocksize)
+                        f2.write(data)
 
     def head(self, path, size=1024):
         """ Get the first ``size`` bytes from file """
@@ -439,7 +580,7 @@ class AbstractFileSystem(object):
     def tail(self, path, size=1024):
         """ Get the last ``size`` bytes from file """
         with self.open(path, 'rb') as f:
-            f.seek(-size, 2)
+            f.seek(max(-size, -f.size), 2)
             return f.read()
 
     def copy(self, path1, path2, **kwargs):
@@ -485,9 +626,9 @@ class AbstractFileSystem(object):
 
     @classmethod
     def _parent(cls, path):
-        path = path.rstrip('/')
+        path = path.rstrip('/').lstrip('/')
         if '/' in path:
-            return path.rsplit('/', 1)[0]
+            return cls.root_marker + path.rsplit('/', 1)[0]
         else:
             return cls.root_marker
 
@@ -513,6 +654,7 @@ class AbstractFileSystem(object):
             Some indication of buffering - this is a value in bytes
         """
         import io
+        path = self._strip_protocol(path)
         if 'b' not in mode:
             mode = mode.replace('t', '') + 'b'
             return io.TextIOWrapper(self.open(path, mode, block_size, **kwargs))
@@ -524,9 +666,18 @@ class AbstractFileSystem(object):
                 self.transaction.files.append(f)
             return f
 
-    def touch(self, path, **kwargs):
-        """ Create empty file, or update timestamp """
-        if not self.exists(path):
+    def touch(self, path, truncate=True, **kwargs):
+        """ Create empty file, or update timestamp
+
+        Parameters
+        ----------
+        path : str
+            file location
+        truncate : bool
+            If True, always set file size to 0; if False, update timestamp and
+            leave file unchanged, if backend allows this
+        """
+        if truncate or not self.exists(path):
             with self.open(path, 'wb', **kwargs):
                 pass
         else:
@@ -577,7 +728,7 @@ class AbstractFileSystem(object):
             size = f.size
             if length is None:
                 length = size
-            if offset + length > size:
+            if size is not None and offset + length > size:
                 length = size - offset
             return read_block(f, offset, length, delimiter)
 
@@ -617,7 +768,8 @@ class Transaction(object):
     """Filesystem transaction write context
 
     Gathers files for deferred commit or discard, so that several write
-    operations can be finalized semi-atomically.
+    operations can be finalized semi-atomically. This works by having this
+    instance as the ``.transaction`` attribute of the given filesystem
     """
 
     def __init__(self, fs):
@@ -630,8 +782,7 @@ class Transaction(object):
         self.files = []
 
     def __enter__(self):
-        """Start a transaction on this FileSystem"""
-        self.fs._intrans = True
+        self.start()
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         """End transaction and commit, if exit is not due to exception"""
@@ -639,8 +790,12 @@ class Transaction(object):
         self.complete(commit=exc_type is None)
         self.fs._intrans = False
 
+    def start(self):
+        """Start a transaction on this FileSystem"""
+        self.fs._intrans = True
+
     def complete(self, commit=True):
-        """Finish transation: commit or discard all deferred files"""
+        """Finish transaction: commit or discard all deferred files"""
         for f in self.files:
             if commit:
                 f.commit()
@@ -650,7 +805,7 @@ class Transaction(object):
         self.fs._intrans = False
 
 
-class AbstractBufferedFile(object):
+class AbstractBufferedFile(io.IOBase):
     """Convenient class to derive from to provide buffering
 
     In the case that the backend does not provide a pythonic file-like object
@@ -661,49 +816,73 @@ class AbstractBufferedFile(object):
     DEFAULT_BLOCK_SIZE = 5 * 2**20
 
     def __init__(self, fs, path, mode='rb', block_size='default',
-                 autocommit=True, **kwargs):
+                 autocommit=True, cache_type='bytes', **kwargs):
         """
         Template for files with buffered reading and writing
 
         Parameters
         ----------
-        fa: instance of FileSystem
+        fs: instance of FileSystem
         path: str
             location in file-system
         mode: str
-            Normal file modes. Currently only 'wb' amd 'rb'.
+            Normal file modes. Currently only 'wb', 'ab' or 'rb'. Some file
+            systems may be read-only, and some may not support append.
         block_size: int
             Buffer size for reading or writing, 'default' for class default
         autocommit: bool
             Whether to write to final destination; may only impact what
             happens when file is being closed.
+        cache_type : str
+            Caching policy in read mode, one of 'none', 'bytes', 'mmap', see
+            the definitions in ``core``.
         kwargs:
             Gets stored as self.kwargs
         """
+        from .core import caches
         self.path = path
         self.fs = fs
         self.mode = mode
         self.blocksize = (self.DEFAULT_BLOCK_SIZE
                           if block_size == 'default' else block_size)
-        self.cache = b""
         self.loc = 0
         self.autocommit = autocommit
         self.end = None
         self.start = None
         self.closed = False
-        self.trim = True
+        self.trim = kwargs.pop('trim', True)
         self.kwargs = kwargs
-        if mode not in {'rb', 'wb'}:
+        if mode not in {'ab', 'rb', 'wb'}:
             raise NotImplementedError('File mode not supported')
         if mode == 'rb':
-            self.details = fs.info(path)
+            if not hasattr(self, 'details'):
+                self.details = fs.info(path)
             self.size = self.details['size']
+            self.cache = caches[cache_type](self.blocksize, self._fetch_range,
+                                            self.size, trim=self.trim)
         else:
             self.buffer = io.BytesIO()
             self.offset = 0
             self.forced = False
             self.location = None
 
+    @property
+    def closed(self):
+        # get around this attr being read-only in IOBase
+        return self._closed
+
+    @closed.setter
+    def closed(self, c):
+        self._closed = c
+
+    def __hash__(self):
+        return self.fs.checksum(self.path)
+
+    def __eq__(self, other):
+        """Files are equal if they have the same checksum, only in read mode"""
+        return (self.mode == 'rb' and other.mode == 'rb'
+                and hash(self) == hash(other))
+
     def commit(self):
         """Move from temp to final destination"""
 
@@ -728,6 +907,7 @@ class AbstractBufferedFile(object):
         whence : {0, 1, 2}
             from start of file, current location or end of file, resp.
         """
+        loc = int(loc)
         if not self.mode == 'rb':
             raise ValueError('Seek only available in read mode')
         if whence == 0:
@@ -796,15 +976,14 @@ class AbstractBufferedFile(object):
             return
 
         if not self.offset:
-            if not force and self.buffer.tell() <= self.blocksize:
+            if not force and self.buffer.tell() < self.blocksize:
                 # Defer write on small block
                 return
             else:
-                # At initialize a multipart upload
+                # Initialize a multipart upload
                 self._initiate_upload()
 
-        if self._upload_chunk(final=force):
-            # reset write buffer
+        if self._upload_chunk(final=force) is not False:
             self.offset += self.buffer.seek(0, 2)
             self.buffer = io.BytesIO()
 
@@ -820,42 +999,12 @@ class AbstractBufferedFile(object):
             This is the last block, so should complete file, if
             self.autocommit is True.
         """
+        # may not yet have been initialized, may neet to call _initialize_upload
 
     def _initiate_upload(self):
         """ Create remote file/upload """
         pass
 
-    def _fetch(self, start, end):
-        """ Get bytes between start and end, if not already in cache
-
-        Will read ahead by blocksize bytes.
-        """
-        if self.start is None and self.end is None:
-            # First read
-            self.start = start
-            self.end = end + self.blocksize
-            self.cache = self._fetch_range(self.start, self.end)
-        if start < self.start:
-            if self.end - end > self.blocksize:
-                self.start = start
-                self.end = end + self.blocksize
-                self.cache = self._fetch_range(self.start, self.end)
-            else:
-                new = self._fetch_range(start, self.start)
-                self.start = start
-                self.cache = new + self.cache
-        if end > self.end:
-            if self.end > self.size:
-                return
-            if end - self.end > self.blocksize:
-                self.start = start
-                self.end = end + self.blocksize
-                self.cache = self._fetch_range(self.start, self.end)
-            else:
-                new = self._fetch_range(self.end, end + self.blocksize)
-                self.end = end + self.blocksize
-                self.cache = self.cache + new
-
     def _fetch_range(self, start, end):
         """Get the specified set of bytes from remote"""
         raise NotImplementedError
@@ -869,23 +1018,86 @@ class AbstractBufferedFile(object):
         length : int (-1)
             Number of bytes to read; if <0, all remaining bytes.
         """
+        length = int(length)
         if self.mode != 'rb':
             raise ValueError('File not in read mode')
         if length < 0:
             length = self.size
         if self.closed:
             raise ValueError('I/O operation on closed file.')
-        self._fetch(self.loc, self.loc + length)
-        out = self.cache[self.loc - self.start:
-                         self.loc - self.start + length]
+        logger.debug("%s read: %i - %i" % (self, self.loc, self.loc + length))
+        out = self.cache._fetch(self.loc, self.loc + length)
         self.loc += len(out)
-        if self.trim:
-            num = (self.loc - self.start) // self.blocksize - 1
-            if num > 0:
-                self.start += self.blocksize * num
-                self.cache = self.cache[self.blocksize * num:]
         return out
 
+    def readinto(self, b):
+        """mirrors builtin file's readinto method
+
+        https://docs.python.org/3/library/io.html#io.RawIOBase.readinto
+        """
+        data = self.read(len(b))
+        b[:len(data)] = data
+        return len(data)
+
+    def readuntil(self, char=b'\n', blocks=None):
+        """Return data between current position and first occurrence of char
+
+        char is included in the output, except if the end of the tile is
+        encountered first.
+
+        Parameters
+        ----------
+        char : bytes
+            Thing to find
+        blocks : None or int
+            How much to read in each go. Defaults to file blocksize - which may
+            mean a new read on every call.
+        """
+        out = []
+        while True:
+            start = self.tell()
+            part = self.read(blocks or self.blocksize)
+            if len(part) == 0:
+                break
+            found = part.find(char)
+            if found > -1:
+                out.append(part[:found + len(char)])
+                self.seek(start + found + len(char))
+                break
+            out.append(part)
+        return b"".join(out)
+
+    def readline(self):
+        """Read until first occurrence of newline character
+
+        Note that, because of character encoding, this is not necessarily a
+        true line ending.
+        """
+        return self.readuntil(b'\n')
+
+    def __next__(self):
+        out = self.readline()
+        if out:
+            return out
+        raise StopIteration
+
+    def __iter__(self):
+        return self
+
+    def readlines(self):
+        """Return all data, split by the newline character"""
+        data = self.read()
+        lines = data.split(b'\n')
+        out = [l + b'\n' for l in lines[:-1]]
+        if data.endswith(b'\n'):
+            return out + [lines[-1] + b'\n']
+        else:
+            return out + [lines[-1]]
+        # return list(self)  ???
+
+    def readinto1(self, b):
+        return self.readinto(b)
+
     def close(self):
         """ Close file
 
@@ -902,14 +1114,13 @@ class AbstractBufferedFile(object):
                 assert self.buffer.tell() == 0
 
             self.fs.invalidate_cache(self.path)
-            if '/' in self.path:
-                # invalidate parent
-                self.fs.invalidate_cache(self.path.split('/', 1)[0])
+            self.fs.invalidate_cache(self.fs._parent(self.path))
+
         self.closed = True
 
     def readable(self):
         """Whether opened for reading"""
-        return self.mode == 'rb'
+        return self.mode == 'rb' and not self.closed
 
     def seekable(self):
         """Whether is seekable (only in read mode)"""
@@ -917,13 +1128,13 @@ class AbstractBufferedFile(object):
 
     def writable(self):
         """Whether opened for writing"""
-        return self.mode in {'wb', 'ab'}
+        return self.mode in {'wb', 'ab'} and not self.closed
 
     def __del__(self):
         self.close()
 
     def __str__(self):
-        return "<File-like object %s, %s>" % (self.fs, self.path)
+        return "<File-like object %s, %s>" % (type(self.fs).__name__, self.path)
 
     __repr__ = __str__
 
